/********************************************************************\
  Name:         scs_1001_lpcryo.c

  Created by:   Ryu Sawada

  Contents:     Application specific (user) part of
                Midas Slow Control Bus protocol
                for SCS-1001 stand alone control unit

                MEG Large Prototype Xe Calorimeter.

\********************************************************************/

// undef for debugging
#define HAVE_EXT

// Number of channels
#define NRELAY               4    // Number of relais
#define NDOUT                4    // Number of digital outputs
#define NDIN                 4    // Number of digital inputs
#define NDAC                 2    // Number of DAC

// MSCB address of SCS-700
#define SCS700_ADDRESS       3
#define SCS520_ADDRESS      10

// Channel assign
// Relay
#define CH_CIRCPUMP          0    // Circulation pump
// DOUT
#define CH_RF_COMPRESSOR     0    // Compressor
#define CH_LN2               1    // LN2 valve
#define CH_RF_HEATER         2    // Cold head heater
#define CH_BH                3    // Bottom heater
// ADC
#define CH_FLOW              0    // Flow meter    (V)
#define CH_PRESSURE          2    // Pressure      (A)
#define CH_RF_TEMP           3    // Cold head     (A)
#define CH_BH_TEMP           4    // Bottom heater (A)

// Do PID control every second. 
#define PERIOD             100    // *10 msec

// PID control parameters
#define RF_PID_Temp_Prop    10.   // Propotional for temperature control mode.
#define RF_PID_Temp_Int      1.   // Integral for temperature control mode.
#define RF_PID_Pres_Prop  5000.   // Propotional for pressure control mode.
#define RF_PID_Pres_Int    500.   // Integral for pressure control mode.
#define BH_PID_Temp_Prop    10.   // Propotional for temperature control mode.
#define BH_PID_Temp_Int      5.   // Integral for temperature control mode.
#define BH_PID_Pres_Prop 10000.   // Propotional for pressure control mode.
#define BH_PID_Pres_Int   5000.   // Integral for pressure control mode.

// Threshold to turn off heater
#define T_Limit            450.   // Turn off heater when temperature is higher.
#define P_Limit              0.07 // Turn off heater when pressure is lower.

// Control mod
#define CONT_MANUAL          0    // Manual control
#define CONT_PRECOOL         1    // Liquefaction mode
#define CONT_LIQUEFY         2    // Pre cooling mode
#define CONT_OPERATION       3    // Stable operation mode
#define CONT_CIRCULATION     4    // Gas phase circulation  mode
#define CONT_RECOVER         5    // Recovering mode

// Control mode of heater
#define HEATER_CONT_NONE     0    // Control is off.
#define HEATER_CONT_TEMP     1    // Control according to temperature.
#define HEATER_CONT_PRES     2    // Control according to pressure.

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include "mscb.h"

extern bit FREEZE_MODE;
extern bit DEBUG_MODE;

char code node_name[]    = "LP-CRYOSTAT";
char code cvs_revision[] = "$Id: scs_1001_lp.c,v 1.11 2005/08/25 18:49:21 sawada Exp $";

// declare number of sub-addresses to framework
unsigned char idata _n_sub_addr = 1;

extern lcd_menu(void);

//---- Port definitions ----

sbit RELAIS0  = P3 ^ 0;
sbit RELAIS1  = P3 ^ 1;
sbit RELAIS2  = P3 ^ 2;
sbit RELAIS3  = P3 ^ 3;

sbit DOUT0    = P3 ^ 7;
sbit DOUT1    = P3 ^ 6;
sbit DOUT2    = P3 ^ 5;
sbit DOUT3    = P3 ^ 4;

sbit SRCLK    = P1 ^ 7;
sbit SRIN     = P1 ^ 6;
sbit SRSTROBE = P1 ^ 5;

//---- Define variable parameters returned to CMD_GET_INFO command ----

// data buffer (mirrored in EEPROM)

struct {
      // Internal Outputs
      unsigned char relais[NRELAY];
      unsigned char dout[NDOUT];
      unsigned char dout_power[NDOUT];
      unsigned char dac[NDAC];

      unsigned char preset_mode;

      // Alarm
      unsigned char alarm;
      unsigned char alarm_on;
      float         alarm_press_up;
      float         alarm_press_low;

      // Inner vessel
      float         pressure;

      // Flow
      unsigned char circpump_on;
      float         flow;

      // Refrigerator
      float         rf_temperature;
      unsigned char rf_heater_power;
      unsigned char rf_heater_control_mode;
      float         rf_temp_demand;
      float         rf_pres_demand;
      float         rf_p_int;

      unsigned char rf_compressor_on;
      unsigned char rf_compressor_control_on;
      float         rf_pres_th_up;
      float         rf_pres_th_low;

      // Bottom heater
      float         bh_temperature;
      unsigned char bh_power;
      unsigned char bh_control_mode;
      float         bh_temp_demand;
      float         bh_pres_demand;
      float         bh_p_int;

      // LN2
      unsigned char ln2_on;
      unsigned char ln2_control_on;
      float         ln2_pres_th_up;
      float         ln2_pres_th_low;

      // Channel assign
      unsigned short pressure_chn;
      unsigned short flow_chn;
      unsigned short rf_temp_chn;
      unsigned short bh_temp_chn;

      // Calirated temperature
      float temperature[8];

      // Internal Inputs
      unsigned char din[NDIN];
      float adc[8];
#ifdef HAVE_EXT
      float scs700[8];
      float scs520[8];
#endif

      // Calibration parameters
      float tofs[8];
      float aofs[8];
      float again[8];
      float dofs[NDAC];
      float dgain[NDAC];

} xdata user_data;

MSCB_INFO_VAR code variables[] = {
   // Relais
   { 1, UNIT_BOOLEAN, 0,         0, MSCBF_HIDDEN | MSCBF_HIDDEN,"Relay0", &user_data.relais[0],                  0, 1, 1 }, // 0
   { 1, UNIT_BOOLEAN, 0,         0, MSCBF_HIDDEN | MSCBF_HIDDEN,"Relay1", &user_data.relais[1],                  0, 1, 1 },
   { 1, UNIT_BOOLEAN, 0,         0, MSCBF_HIDDEN | MSCBF_HIDDEN,"Relay2", &user_data.relais[2],                  0, 1, 1 },
   { 1, UNIT_BOOLEAN, 0,         0, MSCBF_HIDDEN | MSCBF_HIDDEN,"Relay3", &user_data.relais[3],                  0, 1, 1 },

   // Digital out
   { 1, UNIT_BOOLEAN, 0,         0, MSCBF_HIDDEN | MSCBF_HIDDEN,"Dout0",  &user_data.dout[0],                    0, 1, 1 }, // 4
   { 1, UNIT_BOOLEAN, 0,         0, MSCBF_HIDDEN | MSCBF_HIDDEN,"Dout1",  &user_data.dout[1],                    0, 1, 1 },
   { 1, UNIT_BOOLEAN, 0,         0, MSCBF_HIDDEN | MSCBF_HIDDEN,"Dout2",  &user_data.dout[2],                    0, 1, 1 },
   { 1, UNIT_BOOLEAN, 0,         0, MSCBF_HIDDEN | MSCBF_HIDDEN,"Dout3",  &user_data.dout[3],                    0, 1, 1 },

   { 1, UNIT_PERCENT, 0,         0, MSCBF_HIDDEN | MSCBF_HIDDEN,"Dout0P", &user_data.dout_power[0],              0, 100, 1 }, // 8
   { 1, UNIT_PERCENT, 0,         0, MSCBF_HIDDEN | MSCBF_HIDDEN,"Dout1P", &user_data.dout_power[1],              0, 100, 1 },
   { 1, UNIT_PERCENT, 0,         0, MSCBF_HIDDEN | MSCBF_HIDDEN,"Dout2P", &user_data.dout_power[2],              0, 100, 1 },
   { 1, UNIT_PERCENT, 0,         0, MSCBF_HIDDEN | MSCBF_HIDDEN,"Dout3P", &user_data.dout_power[3],              0, 100, 1 },

   { 4, UNIT_VOLT,    0,         0, MSCBF_FLOAT | MSCBF_HIDDEN, "DAC0",   &user_data.dac[0],                   -10, 10, 0.01 }, // 12
   { 4, UNIT_VOLT,    0,         0, MSCBF_FLOAT | MSCBF_HIDDEN, "DAC1",   &user_data.dac[1],                   -10, 10, 0.01 },

   // Preset control
   { 1, UNIT_BYTE,    0,         0, 0,                          "PreSet", &user_data.preset_mode,                0, 6, 1 }, // 14

   // Alarm
   { 1, UNIT_BOOLEAN, 0,         0, 0,                          "Alarm",  &user_data.alarm,                      0, 1, 1 }, // 15
   { 1, UNIT_BOOLEAN, 0,         0, 0,                          "AlamOn", &user_data.alarm_on,                   0, 1, 1 },
   { 4, UNIT_PASCAL , PRFX_MEGA, 0, MSCBF_FLOAT,                "AlPrUT", &user_data.alarm_press_up,             0, 0.3, 0.01 },
   { 4, UNIT_PASCAL , PRFX_MEGA, 0, MSCBF_FLOAT,                "AlPrLT", &user_data.alarm_press_low,            0, 0.3, 0.01 },

   // Inner vessel
   { 4, UNIT_PASCAL , PRFX_MEGA, 0, MSCBF_FLOAT,                "Press",  &user_data.pressure },                            // 19

   // Flow
   { 1, UNIT_BOOLEAN, 0,         0, 0,                          "CiPmOn", &user_data.circpump_on,                0, 1, 1 }, // 20
   { 4, UNIT_PASCAL , PRFX_MEGA, 0, MSCBF_FLOAT ,               "Flow",   &user_data.flow },

   // Refrigerator
   { 4, UNIT_KELVIN , 0,         0, MSCBF_FLOAT ,               "RFTemp", &user_data.rf_temperature },                      // 22
   { 1, UNIT_PERCENT, 0,         0, 0,                          "RFHtPw", &user_data.rf_heater_power,            0, 100, 1},
   { 1, UNIT_BYTE,    0,         0, 0,                          "RFCtMd", &user_data.rf_heater_control_mode,     0, 2, 1 },
   { 4, UNIT_KELVIN , 0,         0, MSCBF_FLOAT ,               "RFTDmd", &user_data.rf_temp_demand,             50,200,0.5 },
   { 4, UNIT_KELVIN , 0,         0, MSCBF_FLOAT ,               "RFPDmd", &user_data.rf_pres_demand,             50,200,0.5 },
   { 4, UNIT_PERCENT, 0,         0, MSCBF_FLOAT ,               "RFPInt", &user_data.rf_p_int,                   0,0,0 },

   { 1, UNIT_BOOLEAN, 0,         0, 0,                          "RFCmOn", &user_data.rf_compressor_on,           0, 1, 1 },
   { 1, UNIT_BOOLEAN, 0,         0, 0,                          "RFCtOn", &user_data.rf_compressor_control_on,   0, 1, 1 },
   { 4, UNIT_PASCAL , PRFX_MEGA, 0, MSCBF_FLOAT ,               "RFCUTh", &user_data.rf_pres_th_up,              0.07, 0.22, 0.005},
   { 4, UNIT_PASCAL , PRFX_MEGA, 0, MSCBF_FLOAT ,               "RFCLTh", &user_data.rf_pres_th_low,             0.07, 0.22, 0.005},

   // Bottom heater
   { 4, UNIT_KELVIN , 0,         0, MSCBF_FLOAT ,               "BHTemp", &user_data.bh_temperature },                     // 32
   { 1, UNIT_PERCENT, 0,         0, 0,                          "BHHtPw", &user_data.bh_power,                   0, 100, 1},
   { 1, UNIT_BYTE,    0,         0, 0,                          "BHCtMd", &user_data.bh_control_mode,            0, 2, 1 },
   { 4, UNIT_KELVIN , 0,         0, MSCBF_FLOAT ,               "BHTDmd", &user_data.bh_temp_demand,             50,200,0.5 },
   { 4, UNIT_KELVIN , 0,         0, MSCBF_FLOAT ,               "BHPDmd", &user_data.bh_pres_demand,             50,200,0.5 },
   { 4, UNIT_PERCENT, 0,         0, MSCBF_FLOAT ,               "RFPInt", &user_data.bh_p_int,                   0,0,0 },

   // LN2
   { 1, UNIT_BOOLEAN, 0,         0, 0,                          "LN2On",  &user_data.ln2_on,                     0, 1, 1 }, //38
   { 1, UNIT_BOOLEAN, 0,         0, 0,                          "LNCtOn", &user_data.ln2_control_on,             0, 1, 1 },
   { 4, UNIT_PASCAL , PRFX_MEGA, 0, MSCBF_FLOAT ,               "LNUpTh", &user_data.ln2_pres_th_up,             0.07, 0.22, 0.005},
   { 4, UNIT_PASCAL , PRFX_MEGA, 0, MSCBF_FLOAT ,               "LNLwTh", &user_data.ln2_pres_th_low,            0.07, 0.22, 0.005},

   // Channel assign (not used.)
   { 2, UNIT_WORD,    0,         0, 0,                          "Resv1",  &user_data.pressure_chn,               0,7,1}, //42
   { 2, UNIT_WORD,    0,         0, 0,                          "Resv2",  &user_data.pressure_chn,               0,7,1},
   { 2, UNIT_WORD,    0,         0, 0,                          "Resv3",  &user_data.rf_temp_chn,                0,7,1},
   { 2, UNIT_WORD,    0,         0, 0,                          "Resv4",  &user_data.bh_temp_chn,                0,7,1},

   // Calibrated temperatures
   { 4, UNIT_CELSIUS, 0,         0, MSCBF_FLOAT,                "TEMP0",  &user_data.temperature[0] }, // 46
   { 4, UNIT_CELSIUS, 0,         0, MSCBF_FLOAT,                "TEMP1",  &user_data.temperature[1] },
   { 4, UNIT_CELSIUS, 0,         0, MSCBF_FLOAT,                "TEMP2",  &user_data.temperature[2] },
   { 4, UNIT_CELSIUS, 0,         0, MSCBF_FLOAT,                "TEMP3",  &user_data.temperature[3] },
   { 4, UNIT_CELSIUS, 0,         0, MSCBF_FLOAT,                "TEMP4",  &user_data.temperature[4] },
   { 4, UNIT_CELSIUS, 0,         0, MSCBF_FLOAT,                "TEMP5",  &user_data.temperature[5] },
   { 4, UNIT_CELSIUS, 0,         0, MSCBF_FLOAT,                "TEMP6",  &user_data.temperature[6] },
   { 4, UNIT_CELSIUS, 0,         0, MSCBF_FLOAT,                "TEMP7",  &user_data.temperature[7] },

   // Internal inputs
   { 1, UNIT_BOOLEAN, 0,         0, MSCBF_HIDDEN | MSCBF_HIDDEN,"Dout0",  &user_data.din[0] }, // 54
   { 1, UNIT_BOOLEAN, 0,         0, MSCBF_HIDDEN | MSCBF_HIDDEN,"Dout1",  &user_data.din[1] },
   { 1, UNIT_BOOLEAN, 0,         0, MSCBF_HIDDEN | MSCBF_HIDDEN,"Dout2",  &user_data.din[2] },
   { 1, UNIT_BOOLEAN, 0,         0, MSCBF_HIDDEN | MSCBF_HIDDEN,"Dout3",  &user_data.din[3] },

   { 4, UNIT_VOLT,    0,         0, MSCBF_FLOAT | MSCBF_HIDDEN, "ADC0",   &user_data.adc[0] }, // 58
   { 4, UNIT_VOLT,    0,         0, MSCBF_FLOAT | MSCBF_HIDDEN, "ADC1",   &user_data.adc[1] },
   { 4, UNIT_VOLT,    0,         0, MSCBF_FLOAT | MSCBF_HIDDEN, "ADC2",   &user_data.adc[2] },
   { 4, UNIT_VOLT,    0,         0, MSCBF_FLOAT | MSCBF_HIDDEN, "ADC3",   &user_data.adc[3] },
   { 4, UNIT_VOLT,    0,         0, MSCBF_FLOAT | MSCBF_HIDDEN, "ADC4",   &user_data.adc[4] },
   { 4, UNIT_VOLT,    0,         0, MSCBF_FLOAT | MSCBF_HIDDEN, "ADC5",   &user_data.adc[5] },
   { 4, UNIT_VOLT,    0,         0, MSCBF_FLOAT | MSCBF_HIDDEN, "ADC6",   &user_data.adc[6] },
   { 4, UNIT_VOLT,    0,         0, MSCBF_FLOAT | MSCBF_HIDDEN, "ADC7",   &user_data.adc[7] },

   // External channels
#ifdef HAVE_EXT
   { 4, UNIT_CELSIUS, 0,         0, MSCBF_FLOAT | MSCBF_REMIN , "S700-0", &user_data.scs700[0] ,                 0,0,0, SCS700_ADDRESS,8}, // 66
   { 4, UNIT_CELSIUS, 0,         0, MSCBF_FLOAT | MSCBF_REMIN , "S700-1", &user_data.scs700[1] ,                 0,0,0, SCS700_ADDRESS,9},
   { 4, UNIT_CELSIUS, 0,         0, MSCBF_FLOAT | MSCBF_REMIN , "S700-2", &user_data.scs700[2] ,                 0,0,0, SCS700_ADDRESS,10},
   { 4, UNIT_CELSIUS, 0,         0, MSCBF_FLOAT | MSCBF_REMIN , "S700-3", &user_data.scs700[3] ,                 0,0,0, SCS700_ADDRESS,11},
   { 4, UNIT_CELSIUS, 0,         0, MSCBF_FLOAT | MSCBF_REMIN , "S700-4", &user_data.scs700[4] ,                 0,0,0, SCS700_ADDRESS,12},
   { 4, UNIT_CELSIUS, 0,         0, MSCBF_FLOAT | MSCBF_REMIN , "S700-5", &user_data.scs700[5] ,                 0,0,0, SCS700_ADDRESS,13},
   { 4, UNIT_CELSIUS, 0,         0, MSCBF_FLOAT | MSCBF_REMIN , "S700-6", &user_data.scs700[6] ,                 0,0,0, SCS700_ADDRESS,14},
   { 4, UNIT_CELSIUS, 0,         0, MSCBF_FLOAT | MSCBF_REMIN , "S700-7", &user_data.scs700[7] ,                 0,0,0, SCS700_ADDRESS,15},

   { 4, UNIT_CELSIUS, 0,         0, MSCBF_FLOAT | MSCBF_REMIN , "S520-0", &user_data.scs520[0] ,                 0,0,0, SCS520_ADDRESS,0}, // 74
   { 4, UNIT_CELSIUS, 0,         0, MSCBF_FLOAT | MSCBF_REMIN , "S520-1", &user_data.scs520[1] ,                 0,0,0, SCS520_ADDRESS,1},
   { 4, UNIT_CELSIUS, 0,         0, MSCBF_FLOAT | MSCBF_REMIN , "S520-2", &user_data.scs520[2] ,                 0,0,0, SCS520_ADDRESS,2},
   { 4, UNIT_CELSIUS, 0,         0, MSCBF_FLOAT | MSCBF_REMIN , "S520-3", &user_data.scs520[3] ,                 0,0,0, SCS520_ADDRESS,3},
   { 4, UNIT_CELSIUS, 0,         0, MSCBF_FLOAT | MSCBF_REMIN , "S520-4", &user_data.scs520[4] ,                 0,0,0, SCS520_ADDRESS,4},
   { 4, UNIT_CELSIUS, 0,         0, MSCBF_FLOAT | MSCBF_REMIN , "S520-5", &user_data.scs520[5] ,                 0,0,0, SCS520_ADDRESS,5},
   { 4, UNIT_CELSIUS, 0,         0, MSCBF_FLOAT | MSCBF_REMIN , "S520-6", &user_data.scs520[6] ,                 0,0,0, SCS520_ADDRESS,6},
   { 4, UNIT_CELSIUS, 0,         0, MSCBF_FLOAT | MSCBF_REMIN , "S520-7", &user_data.scs520[7] ,                 0,0,0, SCS520_ADDRESS,7},
#endif

   // Calibration
   { 4, UNIT_VOLT,    0,         0, MSCBF_FLOAT | MSCBF_HIDDEN, "TOFS0",  &user_data.tofs[0],                    -50, 50, 0.5 }, // 82
   { 4, UNIT_VOLT,    0,         0, MSCBF_FLOAT | MSCBF_HIDDEN, "TOFS1",  &user_data.tofs[1],                    -50, 50, 0.5 },
   { 4, UNIT_VOLT,    0,         0, MSCBF_FLOAT | MSCBF_HIDDEN, "TOFS2",  &user_data.tofs[2],                    -50, 50, 0.5 },
   { 4, UNIT_VOLT,    0,         0, MSCBF_FLOAT | MSCBF_HIDDEN, "TOFS3",  &user_data.tofs[3],                    -50, 50, 0.5 },
   { 4, UNIT_VOLT,    0,         0, MSCBF_FLOAT | MSCBF_HIDDEN, "TOFS4",  &user_data.tofs[4],                    -50, 50, 0.5 },
   { 4, UNIT_VOLT,    0,         0, MSCBF_FLOAT | MSCBF_HIDDEN, "TOFS5",  &user_data.tofs[5],                    -50, 50, 0.5 },
   { 4, UNIT_VOLT,    0,         0, MSCBF_FLOAT | MSCBF_HIDDEN, "TOFS6",  &user_data.tofs[6],                    -50, 50, 0.5 },
   { 4, UNIT_VOLT,    0,         0, MSCBF_FLOAT | MSCBF_HIDDEN, "TOFS7",  &user_data.tofs[7],                    -50, 50, 0.5 },

   { 4, UNIT_VOLT,    0,         0, MSCBF_FLOAT | MSCBF_HIDDEN, "AOFS0",  &user_data.aofs[0],                    -0.1, 0.1, 0.001 }, // 90
   { 4, UNIT_VOLT,    0,         0, MSCBF_FLOAT | MSCBF_HIDDEN, "AOFS1",  &user_data.aofs[1],                    -0.1, 0.1, 0.001 },
   { 4, UNIT_VOLT,    0,         0, MSCBF_FLOAT | MSCBF_HIDDEN, "AOFS2",  &user_data.aofs[2],                    -0.1, 0.1, 0.001 },
   { 4, UNIT_VOLT,    0,         0, MSCBF_FLOAT | MSCBF_HIDDEN, "AOFS3",  &user_data.aofs[3],                    -0.1, 0.1, 0.001 },
   { 4, UNIT_VOLT,    0,         0, MSCBF_FLOAT | MSCBF_HIDDEN, "AOFS4",  &user_data.aofs[4],                    -0.1, 0.1, 0.001 },
   { 4, UNIT_VOLT,    0,         0, MSCBF_FLOAT | MSCBF_HIDDEN, "AOFS5",  &user_data.aofs[5],                    -0.1, 0.1, 0.001 },
   { 4, UNIT_VOLT,    0,         0, MSCBF_FLOAT | MSCBF_HIDDEN, "AOFS6",  &user_data.aofs[6],                    -0.1, 0.1, 0.001 },
   { 4, UNIT_VOLT,    0,         0, MSCBF_FLOAT | MSCBF_HIDDEN, "AOFS7",  &user_data.aofs[7],                    -0.1, 0.1, 0.001 },

   { 4, UNIT_FACTOR,  0,         0, MSCBF_FLOAT | MSCBF_HIDDEN, "AGAIN0", &user_data.again[0],                   0.9, 1.1, 0.001 }, // 98
   { 4, UNIT_FACTOR,  0,         0, MSCBF_FLOAT | MSCBF_HIDDEN, "AGAIN1", &user_data.again[1],                   0.9, 1.1, 0.001 },
   { 4, UNIT_FACTOR,  0,         0, MSCBF_FLOAT | MSCBF_HIDDEN, "AGAIN2", &user_data.again[2],                   0.9, 1.1, 0.001 },
   { 4, UNIT_FACTOR,  0,         0, MSCBF_FLOAT | MSCBF_HIDDEN, "AGAIN3", &user_data.again[3],                   0.9, 1.1, 0.001 },
   { 4, UNIT_FACTOR,  0,         0, MSCBF_FLOAT | MSCBF_HIDDEN, "AGAIN4", &user_data.again[4],                   0.9, 1.1, 0.001 },
   { 4, UNIT_FACTOR,  0,         0, MSCBF_FLOAT | MSCBF_HIDDEN, "AGAIN5", &user_data.again[5],                   0.9, 1.1, 0.001 },
   { 4, UNIT_FACTOR,  0,         0, MSCBF_FLOAT | MSCBF_HIDDEN, "AGAIN6", &user_data.again[6],                   0.9, 1.1, 0.001 },
   { 4, UNIT_FACTOR,  0,         0, MSCBF_FLOAT | MSCBF_HIDDEN, "AGAIN7", &user_data.again[7],                   0.9, 1.1, 0.001 },

   { 4, UNIT_VOLT,    0,         0, MSCBF_FLOAT | MSCBF_HIDDEN, "DOFS0",  &user_data.dofs[0],                    -0.1, 0.1, 0.001 }, // 106
   { 4, UNIT_VOLT,    0,         0, MSCBF_FLOAT | MSCBF_HIDDEN, "DOFS1",  &user_data.dofs[1],                    -0.1, 0.1, 0.001 },

   { 4, UNIT_FACTOR,  0,         0, MSCBF_FLOAT | MSCBF_HIDDEN, "DGAIN0", &user_data.dgain[0],                   0.9, 1.1, 0.001 }, // 108
   { 4, UNIT_FACTOR,  0,         0, MSCBF_FLOAT | MSCBF_HIDDEN, "DGAIN1", &user_data.dgain[1],                   0.9, 1.1, 0.001 },

   { 0 }
};

/********************************************************************\

  Application specific init and inout/output routines

\********************************************************************/

void user_write(unsigned char index) reentrant;

//---- User init function ------------------------------------------

extern SYS_INFO sys_info;

void user_init(unsigned char init)
{
   unsigned char xdata i;
   xdata char str[64];

   SFRPAGE = ADC0_PAGE;
   AMX0CF = 0x00;               // select single ended analog inputs
   ADC0CF = 0xE0;               // 16 system clocks, gain 1
   ADC0CN = 0x80;               // enable ADC
   REF0CN = 0x00;               // use external voltage reference

   SFRPAGE = LEGACY_PAGE;
   REF0CN = 0x02;               // select external voltage reference

   SFRPAGE = DAC0_PAGE;
   DAC0CN = 0x80;               // enable DAC0
   SFRPAGE = DAC1_PAGE;
   DAC1CN = 0x80;               // enable DAC1

   /*
   // open drain(*) /push-pull:
   P0.0 TX1      P1.0 TP6          P2.0 LED1         P3.0 RELAIS0
   P0.1*RX1      P1.1 BACKLIGHT    P2.1 LCD_E        P3.1 RELAIS1
   P0.2 TX2      P1.2 TP5          P2.2 LCD_RW       P3.2 RELAIS2
   P0.3*RX2      P1.3 TP7          P2.3 LCD_RS       P3.3 RELAIS3

   P0.4 EN1      P1.4 WATCHDOG     P2.4 LCD_DB4      P3.4 DOUT0
   P0.5 EN2      P1.5 SRSTROBE     P2.5 LCD_DB5      P3.5 DOUT1
   P0.6 LED2     P1.6*SRIN         P2.6 LCD_DB6      P3.6 DOUT2
   P0.7 BUZZER   P1.7 SRCLK        P2.7 LCD_DB7      P3.7 DOUT3
   */
   SFRPAGE = CONFIG_PAGE;
   P0MDOUT = 0xF5;
   P1MDOUT = 0xBF;
   P2MDOUT = 0xFF;
   P3MDOUT = 0xFF;

   // initial EEPROM value
   if (init) {
      for (i=0 ; i<8 ; i++) {
         user_data.aofs[i] = 0;
         user_data.again[i] = 1;
      }
   }

#ifdef HAVE_EXT
   /* initialize UART1 for SCS_700 and SCS_520 */
   uart_init(1, BD_115200);
#endif
   
   for (i=0 ; i<NRELAY ; i++)
      user_data.relais[i] = 0;

   for (i=0 ; i<NDOUT ; i++)
      user_data.dout[i] = 0;

   for (i=0 ; i<NDAC ; i++) {
      user_data.dac[i] = 0;
      user_data.dofs[i] = 0;
      user_data.dgain[i] = 1;
   }
    
   for (i=0 ; i<8 ; i++) {
      user_data.aofs[i] = 0;
      user_data.again[i] = 1;
   }

   // write digital outputs
   for (i=0 ; i<NDOUT+NRELAY ; i++)
      user_write(i);

   // write DACs
   user_write(12);
   user_write(13);

   // write remote variables
   for (i = 0; variables[i].width; i++)
      if (variables[i].flags & MSCBF_REMOUT)
         send_remote_var(i);

   // display startup screen
   lcd_goto(0, 0);
   for (i=0 ; i<7-strlen(sys_info.node_name)/2 ; i++)
      puts(" ");
   puts("** ");
   puts(sys_info.node_name);
   puts(" **");
   lcd_goto(0, 1);
   printf("  Address:   %04X", sys_info.node_addr);
   lcd_goto(0, 2);
   strcpy(str, cvs_revision + 21);
   *strchr(str, ' ') = 0;
   printf("  Revision:  %s", str);
}

#pragma NOAREGS

//---- User write function -----------------------------------------

void user_write(unsigned char index) reentrant
{
   unsigned short d;

   switch (index) {

      // RELAIS go through inverter
      case 0: RELAIS0 = !user_data.relais[0]; break;
      case 1: RELAIS1 = !user_data.relais[1]; break;
      case 2: RELAIS2 = !user_data.relais[2]; break;
      case 3: RELAIS3 = !user_data.relais[3]; break;
         
      // DOUT go through inverter
      case 4: DOUT0 = !user_data.dout[0]; break;
      case 5: DOUT1 = !user_data.dout[1]; break;
      case 6: DOUT2 = !user_data.dout[2]; break;
      case 7: DOUT3 = !user_data.dout[3]; break;

      case 12:                    // DAC0
         // assume -10V..+10V range
         d = (((user_data.dac[0] + user_data.dofs[0]) * user_data.dgain[0] + 10) / 20) * 0x1000;
         if (d >= 0x1000)
            d = 0x0FFF;
         if (d < 0)
            d = 0;
         
         SFRPAGE = DAC0_PAGE;
         DAC0L = d & 0xFF;
         DAC0H = d >> 8;
         break;

      case 13:                    // DAC1
         // assume -10V..+10V range
         d = (((user_data.dac[1] + user_data.dofs[1]) * user_data.dgain[1] + 10) / 20) * 0x1000;
         if (d >= 0x1000)
            d = 0x0FFF;
         if (d < 0)
            d = 0;

         SFRPAGE = DAC1_PAGE;
         DAC1L = d & 0xFF;
         DAC1H = d >> 8;
         break;
   }
}

//---- User read function ------------------------------------------

unsigned char user_read(unsigned char index)
{
   if (index);
   return 0;
}

//---- User function called vid CMD_USER command -------------------

unsigned char user_func(unsigned char *data_in, unsigned char *data_out)
{
   // echo input data
   data_out[0] = data_in[0];
   data_out[1] = data_in[1];
   return 2;
}

//---- ADC read function -------------------------------------------

void adc_read(channel, float *d)
{
   unsigned long xdata value;
   unsigned int xdata i, n;
   float xdata gvalue;

   SFRPAGE = ADC0_PAGE;
   AMX0SL = channel & 0x0F;
   ADC0CF = 0xE0;               // 16 system clocks, gain 1

   n = 1 << 10;                 // average 1024 times

   value = 0;
   for (i = 0; i < n; i++) {
      DISABLE_INTERRUPTS;

      SFRPAGE = ADC0_PAGE;
      AD0INT = 0;
      AD0BUSY = 1;
      while (!AD0INT);          // wait until conversion ready, does NOT work with ADBUSY!

      ENABLE_INTERRUPTS;

      value += (ADC0L | (ADC0H << 8));
      yield();
   }

   value >>= 6;

   if (channel < 2)
      gvalue = value / 65536.0 * 20 - 10; // +- 10V range
   else
      gvalue = value / 65536.0 * 10; // 0...10V range

   gvalue -= user_data.aofs[channel];
   gvalue *= user_data.again[channel];

   DISABLE_INTERRUPTS;
   *d = gvalue;
   ENABLE_INTERRUPTS;
}

//---- Shift register read function --------------------------------

bit b0, b1, b2, b3;

#define SHIFT SRCLK = 1; delay_us(1); SRCLK = 0; delay_us(1)

void sr_read()
{
   SRCLK = 0;
   SRIN = 1;
   SRSTROBE = 1;
   delay_us(1);
   SRSTROBE = 0;

   user_data.din[2] = SRIN;
   SHIFT;
   user_data.din[3] = SRIN;
   SHIFT;
   user_data.din[0] = SRIN;
   SHIFT;
   user_data.din[1] = SRIN;

   SHIFT;
   b3 = SRIN;
   SHIFT;
   b2 = SRIN;
   SHIFT;
   b1 = SRIN;
   SHIFT;
   b0 = SRIN;
}

//---- Apply power settings as a fraction of a second --------------

unsigned char xdata cycle;
unsigned char xdata ca[NDOUT];

void set_power(void)
{
   static unsigned long xdata on_time, last_time;

   unsigned char xdata i;
   float xdata frac;
   unsigned long xdata expired;

   // only process each 10ms
   if (time() == last_time)
      return;

   last_time = time();

   // turn output off after on-time expired
   for (i = 0; i < NDOUT; i++)
      if (user_data.dout_power[i] < 100) {
         expired = time() - on_time;
         if (expired >= (unsigned long) (user_data.dout_power[i])) {
            frac = user_data.dout_power[i] - (unsigned long) (user_data.dout_power[i]);

            if (frac == 0 || expired >= (unsigned long) (user_data.dout_power[i]) + 1) {
               P3 &= ~(1 << (i+NRELAY));
            } else if (cycle > 0) {
               if ((float) ca[i] / cycle > frac) {
                  P3 &= ~(1 << (i+NRELAY));
               } else {
                  ca[i]++;
               }
            } else {
               ca[i]++;
            }
         }
      }

   // turn all outputs on every second
   if (time() - on_time >= 100) {
      last_time = on_time = time();
      for (i = 0; i < NDOUT; i++)
         if (user_data.dout_power[i] > 0)
            P3 |= (1 << (i+NRELAY));

      cycle = (cycle + 1) % 10;

      if (cycle == 0)
         for (i = 0; i < NDOUT; i++)
            ca[i] = 0;
   }
}

//---- LP specific code -----------------------------------

void set_compressor(unsigned char flag)
{
   user_data.dout[CH_RF_COMPRESSOR] = flag;
   if(flag == 0){
      user_data.dout_power[CH_RF_COMPRESSOR] = 0;
   }
   else{
      user_data.dout_power[CH_RF_COMPRESSOR] = 100;
   }
   user_write(CH_RF_COMPRESSOR + NRELAY);
}

void set_ln2(unsigned char flag)
{
   user_data.dout[CH_LN2] = flag;
   if(flag == 0){
      user_data.dout_power[CH_LN2] = 0;
   }
   else{
      user_data.dout_power[CH_LN2] = 100;
   }
   user_write(CH_LN2 + NRELAY);
}

void set_circpump(unsigned char flag)
{
   user_data.dout[CH_CIRCPUMP] = flag;
   if(flag == 0){
      user_data.dout_power[CH_CIRCPUMP] = 0;
   }
   else{
      user_data.dout_power[CH_CIRCPUMP] = 100;
   }
   user_write(CH_CIRCPUMP);
}

void check_th(float* min,float* max){
   float xdata tmp;
   if(*min > *max){
      tmp = *max;
      *max = *min;
      *min = tmp;
   }
}

void do_valve_control()
{
   // return if control is off
   if( !user_data.ln2_control_on )
      return;

   check_th(&user_data.ln2_pres_th_low,&user_data.ln2_pres_th_up);
   if( user_data.ln2_pres_th_up < user_data.pressure){
      user_data.ln2_on = 1;
   }
   if( user_data.pressure < user_data.ln2_pres_th_low ){
      user_data.ln2_on = 0;
   }
}

void do_compressor_control()
{
   // return if control is off
   if( !user_data.rf_compressor_control_on )
      return;

   check_th(&user_data.rf_pres_th_low,&user_data.rf_pres_th_up);
   if( user_data.rf_pres_th_up < user_data.pressure){
      user_data.rf_compressor_on = 1;
   }
   if( user_data.pressure < user_data.rf_pres_th_low ){
      user_data.rf_compressor_on = 0;
   }
}

void do_circpump_control()
{
// to be implemented
/*
   // return if control is off
   if( !user_data.citcpump_control_on )
      return;
   check_th(&user_data.ln2_pres_th_low,&user_data.ln2_pres_th_up);
   if( user_data.ln2_pres_th_up < user_data.pressure){
      user_data.ln2_on = 1;
   }
   if( user_data.pressure < user_data.ln2_pres_th_low ){
      user_data.ln2_on = 0;
   }
*/
}

void do_rf_heater_control(void)
{
   float xdata p;
   static unsigned long xdata t;

   // once every specified period
   if (time() > t + PERIOD) {
      t = time();

      switch(user_data.rf_heater_control_mode) {
         case HEATER_CONT_NONE:
            return;

         case HEATER_CONT_TEMP:
            // integral part
            user_data.rf_p_int+=
               (user_data.rf_temp_demand - user_data.rf_temperature) * RF_PID_Temp_Int;
            if (user_data.rf_p_int < 0)
               user_data.rf_p_int = 0;
            if (user_data.rf_p_int> 100)
               user_data.rf_p_int = 100;
            p = user_data.rf_p_int;

            // proportional part
            p += (user_data.rf_temp_demand - user_data.rf_temperature) * RF_PID_Temp_Prop;

            if (p < 0)
               p = 0;
            if (p > 100)
               p = 100;

            user_data.rf_heater_power = ((int) (p * 10 + 0.5)) / 10.0;
            break;

         case HEATER_CONT_PRES:
            // integral part
            user_data.rf_p_int+=
               (user_data.rf_pres_demand - user_data.pressure) * RF_PID_Pres_Int;
            if (user_data.rf_p_int < 0)
               user_data.rf_p_int = 0;
            if (user_data.rf_p_int> 100)
               user_data.rf_p_int = 100;
            p = user_data.rf_p_int;

            // proportional part
            p += (user_data.rf_pres_demand - user_data.pressure) * RF_PID_Pres_Prop;

            if (p < 0)
               p = 0;
            if (p > 100)
               p = 100;

            user_data.rf_heater_power = ((int) (p * 10 + 0.5)) / 10.0;
            break;
      }
   }
}

void do_bh_control(void)
{
   float xdata p;
   static unsigned long xdata t;

   // once every specified period
   if (time() > t + PERIOD) {
      t = time();

      switch(user_data.bh_control_mode) {
         case HEATER_CONT_NONE:
            return;

         case HEATER_CONT_TEMP:
            // integral part
            user_data.bh_p_int+=
               (user_data.bh_temp_demand - user_data.bh_temperature) * BH_PID_Temp_Int;
            if (user_data.bh_p_int < 0)
               user_data.bh_p_int = 0;
            if (user_data.bh_p_int> 100)
               user_data.bh_p_int = 100;
            p = user_data.bh_p_int;

            // proportional part
            p += (user_data.bh_temp_demand - user_data.bh_temperature) * BH_PID_Temp_Prop;

            if (p < 0)
               p = 0;
            if (p > 100)
               p = 100;

            user_data.bh_power = ((int) (p * 10 + 0.5)) / 10.0;
            break;

         case HEATER_CONT_PRES:
            // integral part
            user_data.bh_p_int+=
               (user_data.bh_pres_demand - user_data.pressure) * BH_PID_Pres_Int;
            if (user_data.bh_p_int < 0)
               user_data.bh_p_int = 0;
            if (user_data.bh_p_int> 100)
               user_data.bh_p_int = 100;
            p = user_data.bh_p_int;

            // proportional part
            p += (user_data.bh_pres_demand - user_data.pressure) * BH_PID_Pres_Prop;

            if (p < 0)
               p = 0;
            if (p > 100)
               p = 100;

            user_data.bh_power = ((int) (p * 10 + 0.5)) / 10.0;
            break;
      }
   }
}

unsigned char application_display(bit init)
{
   static bit b0_old = 0, b1_old = 0, b2_old = 0, b3_old = 0;

   if (init)
      lcd_clear();

   // display control mode
   lcd_goto(0,0);
   switch(user_data.preset_mode){
     case CONT_MANUAL: 
      printf("Manual Control     ");
      break;
     case CONT_PRECOOL:
      printf("Pre-Cooling        ");
      break;
     case CONT_LIQUEFY:
      printf("Liquefying         ");
      break;
     case CONT_OPERATION:
      printf("Stable Operation   ");
      break;
     case CONT_CIRCULATION:
      printf("GAS Circulation    ");
      break;
     case CONT_RECOVER:
      printf("Recovering         ");
      break;
   };

   // display pressure
   lcd_goto(0, 1);
   if(user_data.rf_heater_control_mode == HEATER_CONT_PRES ||
      user_data.bh_control_mode == HEATER_CONT_PRES)
      printf("IV:* %10.2E MPa", user_data.pressure);
   else
      printf("IV:  %10.2E MPa", user_data.pressure);

   // display cold head temperature
   lcd_goto(0, 2);
   if(user_data.rf_heater_control_mode == HEATER_CONT_TEMP ||
      user_data.bh_control_mode == HEATER_CONT_TEMP)
      printf("CH:* %10.2E K  ", user_data.rf_temperature);
   else
      printf("CH:  %10.2E K  ", user_data.rf_temperature);

   lcd_goto(0, 3);
   if(user_data.alarm){
      printf("OK");
      if (b0 && !b0_old)
         user_data.alarm = 0;
   }
   else{
      printf("");
   }

   lcd_goto(16, 3);
   printf("MENU");

   /*
     lcd_goto(0, 3);
     printf(user_data.station_on ? "OFF" : "ON ");

     lcd_goto(4, 3);
     printf(user_data.valve_locked ? "UNLOCK" : " LOCK ");
   */
   // toggle main switch with button 0
   /*
     if (b0 && !b0_old)
     user_data.station_on = !user_data.station_on;

     // toggle locking with button 1
     if (b1 && !b1_old)
     user_data.valve_locked = !user_data.valve_locked;
   */

   // enter menu on release of button 3
   if (!init)
      if (!b3 && b3_old)
         return 1;

   b0_old = b0;
   b1_old = b1;
   b2_old = b2;
   b3_old = b3;

   return 0;
}

void call_preset(void)
{
   switch(user_data.preset_mode){
      case CONT_MANUAL:
         break;
      case CONT_PRECOOL:
         user_data.alarm_on                 = 1;
         user_data.alarm_press_up           = 0.13;
         user_data.alarm_press_low          = 0.095;
         user_data.circpump_on              = 0;
         user_data.rf_heater_control_mode   = 2;
         user_data.rf_pres_demand           = 0.11;
         user_data.rf_compressor_control_on = 1;
         user_data.rf_pres_th_up            = 0.12;
         user_data.rf_pres_th_low           = 0.10;
         user_data.bh_power                 = 0;
         user_data.bh_control_mode          = 0;
         user_data.ln2_control_on           = 1;
         user_data.ln2_pres_th_up           = 0.12;
         user_data.ln2_pres_th_low          = 0.105;
       break;
      case CONT_LIQUEFY:
         user_data.alarm_on                 = 1;
         user_data.alarm_press_up           = 0.15;
         user_data.alarm_press_low          = 0.10;
         user_data.circpump_on              = 0;
         user_data.rf_heater_control_mode   = 2;
         user_data.rf_pres_demand           = 0.13;
         user_data.rf_compressor_control_on = 1;
         user_data.rf_pres_th_up            = 0.14;
         user_data.rf_pres_th_low           = 0.105;
         user_data.bh_power                 = 0;
         user_data.bh_control_mode          = 0;
         user_data.ln2_control_on           = 1;
         user_data.ln2_pres_th_up           = 0.135;
         user_data.ln2_pres_th_low          = 0.125;
       break;
      case CONT_OPERATION:
         user_data.alarm_on                 = 1;
         user_data.alarm_press_up           = 0.15;
         user_data.alarm_press_low          = 0.10;
         user_data.circpump_on              = 0;
         user_data.rf_heater_control_mode   = 2;
         user_data.rf_pres_demand           = 0.13;
         user_data.rf_compressor_control_on = 1;
         user_data.rf_pres_th_up            = 0.14;
         user_data.rf_pres_th_low           = 0.105;
         user_data.bh_power                 = 0;
         user_data.bh_control_mode          = 0;
         user_data.ln2_control_on           = 1;
         user_data.ln2_pres_th_up           = 0.135;
         user_data.ln2_pres_th_low          = 0.125;
         break;
      case CONT_CIRCULATION:
         user_data.alarm_on                 = 1;
         user_data.alarm_press_up           = 0.15;
         user_data.alarm_press_low          = 0.10;
         user_data.circpump_on              = 1   ;
         user_data.rf_heater_control_mode   = 2   ;
         user_data.rf_pres_demand           = 0.13;
         user_data.rf_compressor_control_on = 1   ;
         user_data.rf_pres_th_up            = 0.14;
         user_data.rf_pres_th_low           = 0.105;
         user_data.bh_power                 = 0;
         user_data.bh_control_mode          = 0;
         user_data.ln2_control_on           = 1;
         user_data.ln2_pres_th_up           = 0.135;
         user_data.ln2_pres_th_low          = 0.125;
         break;
      case CONT_RECOVER:
         user_data.alarm_on                 = 1;
         user_data.alarm_press_up           = 0.19;
         user_data.alarm_press_low          = 0;
         user_data.circpump_on              = 0   ;
         user_data.rf_heater_control_mode   = 0   ;
         user_data.rf_heater_power          = 0   ;
         user_data.rf_compressor_on         = 0;
         user_data.rf_compressor_control_on = 0   ;
         user_data.bh_control_mode          = 2;
         user_data.bh_pres_demand           = 0.18;
         user_data.ln2_control_on           = 1;
         user_data.ln2_pres_th_up           = 0.185;
         user_data.ln2_pres_th_low          = 0.175;
         break;
   };
   return;
}

void do_alarm(void){
   if(user_data.alarm_on == 0 ||
      user_data.alarm != 0)
      return;

   if(user_data.alarm_press_up !=0 &&
      user_data.alarm_press_up < user_data.pressure)
      user_data.alarm = 1;

   if(user_data.alarm_press_low !=0 &&
      user_data.pressure < user_data.alarm_press_low)
      user_data.alarm = 1;
}

//---- User loop function ------------------------------------------

void user_loop(void)
{
   static unsigned char xdata adc_chn = 0;
   static bit rf_compressor_on_old = 0, ln2_on_old = 0, circpump_on_old = 0;
#ifndef HAVE_EXT
   float xdata t;
#endif
 
   // read one ADC channel
   adc_read(adc_chn, &user_data.adc[adc_chn]);

#ifdef HAVE_EXT
   // apply temperature offset
   // (If possible, correcting registance instead of temperature is better. becuase T is not propotional to R.)
   user_data.temperature[adc_chn] = user_data.scs700[adc_chn] + user_data.tofs[adc_chn];
#else
   if(adc_chn == CH_RF_TEMP ||
      adc_chn == CH_BH_TEMP){
      t = user_data.adc[adc_chn];
      
      // convert to Ohms (1mA excitation)
      t = t / 0.001;
      
      // convert to Kelvin, coefficients obtained from table fit (www.lakeshore.com)
      t = 5.232935E-7 * t * t * t + 0.0009 * t * t + 2.357 * t + 28.288;
      
/*
      // convert to Celsius
      t -= 273.15;
*/

      // apply temperature offset
      // (If possible, correcting registance instead of temperature is better. becuase T is not propotional to R.)
      t += user_data.tofs[adc_chn];
      
      user_data.temperature[adc_chn] = t;
   }
#endif

   // set temperature of cold head.
   if(adc_chn == CH_RF_TEMP)
      user_data.rf_temperature = user_data.temperature[adc_chn];

   // set temperature of bottom heater
   if(adc_chn == CH_BH_TEMP)
      user_data.bh_temperature = user_data.temperature[adc_chn];

   // convert voltage to pressure.
   if(adc_chn == CH_PRESSURE){
#ifdef HAVE_EXT
      user_data.pressure = user_data.scs520[adc_chn]*0.625 - 0.2215;
#else
      user_data.pressure = user_data.adc[adc_chn]*0.625 - 0.2215;
#endif
   }

   // convert voltage to flow rate.
   if(adc_chn == CH_FLOW){
#ifdef HAVE_EXT
      user_data.flow = user_data.scs520[adc_chn]*4;
#else
      user_data.flow = user_data.adc[adc_chn]*4;
#endif
   }

   adc_chn = (adc_chn + 1) % 8;

   // read buttons and digital input
   sr_read();

   // call preset
   call_preset();

   // check if alarm should be on.
   do_alarm();

   // Refrigerator
   do_rf_heater_control();
 
   // Bottom heater
   do_bh_control();

   // Set heater power
   set_power();

   // control LN2 vale
   do_valve_control();
   if (user_data.ln2_on && !ln2_on_old) {
      set_ln2(1);
   }
   if (!user_data.ln2_on && ln2_on_old) {
      set_ln2(0);
   }
   ln2_on_old = user_data.ln2_on;

   // control compressor
   do_compressor_control();
   if (user_data.rf_compressor_on && !rf_compressor_on_old) {
      set_compressor(1);
   }
   if (!user_data.rf_compressor_on && rf_compressor_on_old) {
      set_compressor(0);
   }
   rf_compressor_on_old = user_data.rf_compressor_on;

   // control circulation pump
   do_circpump_control();
   if (user_data.circpump_on && !circpump_on_old) {
      set_circpump(1);
   }
   if (!user_data.circpump_on && circpump_on_old) {
      set_circpump(0);
   }
   circpump_on_old = user_data.circpump_on;

   // mangae menu on LCD display
   lcd_menu();
}
